// Jimenez's conditional branch predictor
// Registered paper #3 "OH-SNAP: Optimized Hybrid Scaled Neural Analog Predictor"
// Authors: Daniel A. Jimenez (The University of Texas at San Antonio)

#include <stdio.h>
#include <stdlib.h>
#include <cassert>
#include <string.h>
#include <inttypes.h>

using namespace std;

#include "cbp3_def.h"
#include "cbp3_framework.h"

// we keep a queue with in-flight branch state information in it.  the state is:
// 1.  the index in the path/pattern history arrays for this branch
// 2.  the local pattern history for this branch
// 3.  the prediction : taken or not taken

int queue_head = 0, queue_tail = 0;

// up to 128 branches might be in flight; we need to remember state for
// all of them so we can update them in the retire stage; we keep one more
// to simplify maintenance of the queue

#define MAX_INFLIGHT_BRANCHES	129

class oh_snap_update {
public:
	int ring_idx;		// 9-bit index into the circular buffer of histories
	unsigned int lh;	// 17-bit local history
	bool _prediction;	// the prediction for this branch

	// accessor/mutator methods

	void prediction (bool p) { _prediction = p; }
	bool prediction (void) { return _prediction; }
};

oh_snap_update
	inflight_branch_queue[MAX_INFLIGHT_BRANCHES];	// queue of update state entries

#define WEIGHT_BITS	7
#define MAX_WEIGHT 	((1 << (WEIGHT_BITS - 1)) -1)
#define MIN_WEIGHT 	(- MAX_WEIGHT - 1)

typedef char byte;

// branch cache entry

struct bc_entry {
	unsigned int 
		tag;		// the tag from the PC for this branch 
	bool 
		ever_taken, 	// true if the branch was ever taken
		ever_nottaken;	// true if the branch was ever not taken
	byte 
		bias;		// perceptron bias weight for this branch
};

int nents[] = { 760, 669, 597, 539, 491, 451, 417, 388, 363, 341, 321, 303, 288, 273, 261, 249, 238, 228, 219, 211, 203, 196, 189, 183, 177, 172, 167, 162, 157, 153, 149, 145, 141, 138, 134, 131, 128, 125, 122, 120, 117, 115, 112, 110, 108, 106, 104, 102, 100, 98, 97, 95, 93, 92, 90, 89, 87, 86, 85, 83, 82, 81, 80, 79, 78, 76, 75, 74, 73, 72, 71, 71, 70, 69, 68, 67, 66, 66, 65, 64, 63, 63, 62, 61, 60, 60, 59 };

// class for the main predictor code

class oh_snap {
private:
	int	
		num_entries,		// number of blocks in a column
		history_length,		// global history length
		local_history_length,	// local history length
		num_local_histories,	// number of local history vectors to keep
		tc,			// counter for threshold adaptivity
		num_coefficients_vectors, // number of coefficient vectors
		abits,			// number of address bits in hash
		block_size,		// size of a block
		bc_sets,		// number of sets in branch cache
		bc_assoc,		// associativity of branch cache
		bc_tagbits,		// partial tag bits for branch cache
		history_ring_size,	// size of branch history buffer
		spec_history_ring_idx,	// current speculative index into branch history buffer
		num_local_entries, 	// number of local perceptrons
		num_thetas,		// number of theta values
		coefficient_bits;	// number of bits in one coefficient

	unsigned int 
		hl1, 			// global history length for other predictor
		hl2,			// local history length for other predictor
		hl1_mask, 		// mask and shift values to help quickly
		hl2_mask, 		// find index into global and local predictor tables
		hl1_shift, 
		hl2_shift;

	double
		threshold,		// fraction of theta below which we use the other predictor
		factor,			// factor by which to increase/decrease coefficients
		local_coefficient,	// factor by which to multiply local perceptron component
		theta2;			// fudge factor by which to multiply theta when deciding whether to train

	int *theta;			// training thresholds


// table-based predictors have 2**11 = 2048 entries

#define LG_HYBRID_SIZE	11
#define HYBRID_SIZE	(1<<LG_HYBRID_SIZE)

	byte gshare[HYBRID_SIZE];	// gshare predictor: two-bit saturating counters
	bool local[HYBRID_SIZE];	// local predictor: single bits

	bc_entry **bc;	// branch cache pointer

	// number of entries in k'th column; this array does not count
	// against the hardware budget because it contains run-time constants
	// that would be represented as the sizes of memories in an actual
	// implementation

	byte	**weights;		// array of weights vectors

	bool	*history_ring;		// branch outcome history, managed as a circular buffer

	unsigned short int
		*address_ring;		// branch address history, managed as a circular buffer

	unsigned int **coefficients;	// coefficients vectors

	unsigned int *local_histories;	// local history vectors (as bit patterns in unsigned ints)

	byte **local_weights; // local perceptrons
	
	// rolling my own memory allocator

#define HEAP_SIZE	200000

	unsigned char *dan_bump, dan_heap[HEAP_SIZE];

        unsigned char *dan_malloc (unsigned int size) {
                unsigned char *r = dan_bump;
                // 16-byte align
                while (size & 0xf) size++;
                dan_bump += size;
                if (dan_bump >= &dan_heap[HEAP_SIZE]) {
                        fprintf (stderr, "heap too small!!\n");
                }
		memset (r, 0, size);
                return r;
        }

	// return an array of weights vectors of the right sizes

	byte **make_weights (void) {

		// each "column" will be an array of block_size x num_entries bytes 

		int num_columns = history_length / block_size;
		while (num_columns * block_size < history_length) num_columns++;
		byte **v = (byte **) dan_malloc (num_columns * sizeof (byte *));
		for (int i=0; i<num_columns; i++) {
			int n = nents[i] * block_size;
			v[i] = (byte *) dan_malloc (n);
			memset (v[i], 0, n);
		}
		return v;
	}

	// look up an entry in the branch cache; returns a pointer
	// to a (possibly newly allocated) matching entry

	bc_entry *lookup (unsigned int pc) {

		// find set index
		int s = pc % bc_sets;

		// find tag

		pc /= bc_sets;
		pc &= (1<<bc_tagbits)-1;

		// get pointer to right set

		bc_entry *set = bc[s];

		// search for a matching tag

		int i;
		for (i=0; i<bc_assoc; i++) if (set[i].tag == pc) break;

		// if no matching tag, allocate a new entry

		if (i == bc_assoc) {
			// replace LRU victim
			i = bc_assoc-1;
			set[i].tag = pc;
			set[i].ever_taken = false;
			set[i].ever_nottaken = false;
			set[i].bias = -2; // weakly not taken bias
		}
		assert (i < bc_assoc);

		// bring i'th entry up to MRU position

		bc_entry t = set[i];
		for (int w=i-1; w>=0; w--) set[w+1] = set[w];
		set[0] = t;

		// return address of the MRU entry

		return &set[0];
	}

#define MAX_COEFFICIENT	64.0
#define MIN_COEFFICIENT	0.35

	// get the j'th coefficient in the i'th coefficient vector modulo number of vectors

	double get_coeff (int i, int j) {
		i %= num_coefficients_vectors;
		return (coefficients[i][j] / (double) ((1<<coefficient_bits)-1)) * MAX_COEFFICIENT;
	}

	// set the j'th coefficient in the i'th coefficient vector modulo number of vectors to d

	void set_coeff (int i, int j, double d) {
		if (d > MAX_COEFFICIENT) d = MAX_COEFFICIENT;
		if (d < MIN_COEFFICIENT) d = MIN_COEFFICIENT;
		d /= MAX_COEFFICIENT;
		i %= num_coefficients_vectors;
		coefficients[i][j] = (unsigned int) (d * (double) ((1<<coefficient_bits)-1));
		if (coefficients[i][j] == 0) coefficients[i][j] = 1;
	}

public:

	// constructor

	oh_snap (	
		int bs = 3,		// block size in weights
		int ne = 512, 		// number of weights vectors
		int hl = 256,		// global history length
		int lhl = 17,		// local history length,
		int nlh = 307,		// number of local histories
		int nle = 125,		// number of local perceptrons
		int ab = 9,		// # of bits of path address to use
		int bcs = 83,		// branch cache number of sets
		int bca = 107,		// branch cache associativity
		int bct = 10,		// branch cache tag size
		int it = 1614,		// initial theta
		double it2 = 1.0346,	// initial theta2
		double th = 0.1588,	// fraction of theta below which conventional predictor takes over
		double fa = 1.000673,	// factor by which to increase/decrease coefficients
		double fu = 6.413, 	// coefficient for local history
		double A = 0.041, 	// parameter for initial coefficients curve
		double B = 0.00337,	// parameter for initial coefficients curve
		int nc = 4,		// number of coefficients curves
		int cb = 24,		// number of bits in each coefficient
		double mm = 1.66,	// minimum initial coefficient value
		int h1 = 10,		// other predictor global history length
		int h2 = 11,		// other predictor local history length
		int nt = 26 		// number of thetas
		) {

		num_thetas = nt;
		hl1 = h1;
		hl2 = h2;
		hl1_mask = (1<<hl1)-1;
		hl2_mask = (1<<hl2)-1;
		hl1_shift = LG_HYBRID_SIZE - hl1;
		hl2_shift = LG_HYBRID_SIZE - hl2;
		local_coefficient = fu;
		local_history_length = lhl;
		num_local_histories = nlh;
		num_coefficients_vectors = nc;
		num_local_entries = nle;
		threshold = th;
		factor = fa;
		coefficient_bits = cb;
		history_length = hl;
		block_size = bs;
		abits = ab;
		num_entries = ne;
		tc = 0;
		theta2 = it2;
		bc_sets = bcs;
		bc_assoc = bca;
		bc_tagbits = bct;

		// initialize from params

		dan_bump = &dan_heap[0];
		memset (dan_heap, 0, sizeof (dan_heap));

		local_histories = (unsigned int *) dan_malloc (num_local_histories * sizeof (unsigned int));
		local_weights = (byte **) dan_malloc (sizeof (byte *) * num_local_entries);
		for (int i=0; i<num_local_entries; i++) {
			local_weights[i] = (byte *) dan_malloc (local_history_length);
		}
		theta = (int *) dan_malloc (num_thetas * sizeof (int));
		for (int i=0; i<num_thetas; i++) {
			theta[i] = it;	
		}

		// make branch cache

		bc = (bc_entry **) dan_malloc (bc_sets * sizeof (bc_entry *));
		for (int s=0; s<bc_sets; s++)
			bc[s] = (bc_entry *) dan_malloc (sizeof (bc_entry) * bc_assoc);

		// make array that says how many entries are allocated to
		// a particular range of history; this array does not count
		// against the hardware budget because it is a runtime constant
		// that would be represented as different sized memories in a
		// real implementation

#if 0
		nents = (int *) dan_malloc (sizeof (int) * (history_length / block_size + 1));
		for (i=0; i<7; i++) nents[i] = 512;
		for (i=7; i<13; i++) nents[i] = 256;
		for (i=13; i<history_length/block_size+1; i++) nents[i] = 128;
#endif

		// the history length must be a multiple of block_size

		while (history_length % block_size) history_length++;

		// allocate and compute global coefficients

		coefficients = (unsigned int **) dan_malloc (num_coefficients_vectors * sizeof (unsigned int *));
		for (int j=0; j<num_coefficients_vectors; j++) {
			coefficients[j] = (unsigned int *) dan_malloc (sizeof (unsigned int) * (history_length + 1));
			for (int i=0; i<=hl; i++) {
				double d = 1.0 / (A + B * i);
				if (d < mm) d = mm;
				set_coeff (j, i, d);
			}
		}

		// initialize perceptron weights

		weights = make_weights ();

		// set up circular buffers for global path and pattern histories

		history_ring_size = history_length + 1 + MAX_INFLIGHT_BRANCHES;
		history_ring = (bool *) dan_malloc (history_ring_size);
		address_ring = (unsigned short int *) dan_malloc (history_ring_size * sizeof (unsigned short int));
		spec_history_ring_idx = 0;
		memset (inflight_branch_queue, 0, sizeof (inflight_branch_queue));

		// initialize hybrid predictor tables

		for (int i=0; i<HYBRID_SIZE; i++) {
			gshare[i] = 1; // weakly not taken
			local[i] = false; // not taken
		}
	}

	// shift a branch outcome and lower bits of pc into the circular buffer

	void shift_history_ring (int *ring_idx, bool t, unsigned int address) {
		if (*ring_idx > 0) 
			(*ring_idx)--;
		else
			*ring_idx = history_ring_size - 1;

		// xor the history bit with an address bit to reduce bias
		// for improved learning

		history_ring[*ring_idx] = t ^ !!(address & 16);
		address = (address >> 1) ^ t;
		address_ring[*ring_idx] = (unsigned short int) address & ((1<<abits)-1);
	}

	// get the i'th bit of history from the circular buffer

	bool index_history (int ring_idx, int i) {
		int idx = ring_idx + i;
		bool bit = history_ring[idx % history_ring_size];
		return bit;
	}

	// get the i'th (piece of) address from the circular buffer

	unsigned int index_address (int ring_idx, int i) {
		int idx = ring_idx + i;
		return address_ring[idx % history_ring_size];
	}

	// shuffle bits in address history to provide a hash
	// of several addresses

	unsigned int shuffle (int ring_idx, int v, int n) {
		int	i, j, x, mask, count;

		mask = 1;
		i = (v / block_size) % n;
		count = n;
		x = 0;
		for (j=0; j<abits; j++) {
			x <<= 1;
			x |= !!(mask & index_address (ring_idx, v+i));
			if (!--count) {
				mask <<= 1;
				count = n;
			}
			if (++i == n) i = 0;
		}
		return x;
	}

	// produce an index into the weights array for a given address
	// history and branch PC

	unsigned int index_entries (unsigned int x, unsigned int ring_idx, unsigned int v) {
		int n = block_size;
		assert (n);
		unsigned int z = shuffle (ring_idx, v, n) % (1<<abits);
		z ^= x;
		return z % nents[v/block_size];
	}

	// compute the perceptron output

	double compute_output (unsigned int pc, unsigned int lh, int ring_idx, bc_entry *e) {
		int	i;
		double
			sum_pos, 	// sum of global partial products
			sum, 		// bias weight
			sum_local;	// sum of local partial products

		// lookup table to easily go from binary to bipolar

		static int lut[2] = { -1, 1 };

		// get bias weight times bias coefficient for this pc

		sum = e->bias * get_coeff (pc, 0);

		// global sum starts at 0

		sum_pos = 0;

		// sum local history component

		sum_local = 0;
		byte *lw = local_weights[pc % num_local_entries];
		for (i=0; i<local_history_length; i++) {
			bool bit = !!(lh & (1<<i));
			int h = lut[bit];
			sum_local += h * lw[i];
		}

		// multiply by local fudge factor

		sum_local *= local_coefficient;

		// sum global history component

		for (i=0; i<history_length; i+=block_size) {
			int idx = index_entries(pc, ring_idx, i);
			byte *bl = &weights[i/block_size][idx * block_size];
			for (int j=0; j<block_size; j++) {
				int h = lut[index_history(ring_idx,i+j)];
				sum_pos += h * bl[j] * get_coeff (pc, i+j+1);
			} 
		}

		// return sum of all components

		return sum + sum_pos + sum_local;
	}

	// absolute value

	double absolute_value (double x) {
		if (x < 0) return -x;
		return x;
	}

	// saturating increment/decrement for a weight

	void saturating_increment (byte *c, bool inc) {
		if (inc) {
			if (*c < MAX_WEIGHT) (*c)++; 
		} else {
			if (*c > MIN_WEIGHT) (*c)--;
		}
	}

	// train the perceptron formed by looking at one weight from each
	// of the addresses in v[]

	void train (unsigned int pc, unsigned int lh, int ring_idx, bc_entry *e, bool taken, bool correct) {
		int	i, tpc = pc % num_thetas;
		byte *c;

		// lookup table to make conversion from binary to bipolar easy

		static int lut[2] = { -1, 1 };

		// get the absolute value of the perceptron output

		double b = compute_output (pc, lh, ring_idx, e);
		double a = absolute_value (b);

		// dynamic threshold updating per Seznec
		// we use multiple thresholds trained independently

		if (!correct) {
			tc++;
			if (tc >= 1) {
				theta[tpc]++;
				if (theta[tpc] >= 4095) theta[tpc] = 4095;
				tc = 0;
			}
		}
		if (correct && a < theta[tpc]) {
			tc--;
			if (tc <= -1) {
				theta[tpc]--;
				if (theta[tpc] < 1) theta[tpc] = 1;
				tc = 0;
			}
		}

		// if the prediction is correct and the output is at least theta
		// multiplied by a fudge factor then we do not need to train

		if (correct && a >= (theta[tpc] * theta2)) return;

		// train the bias coefficient; if the bias weight's prediction is
		// correct, increase the coefficient, else decrease it

		if ((e->bias>=0) == taken)
			set_coeff (pc, 0, get_coeff (pc, 0) * factor);
		else
			set_coeff (pc, 0, get_coeff (pc, 0) / factor);

		// train the bias weight

		c = &e->bias;
		saturating_increment (c, taken);

		// train the local weights; they do not have coefficients

		c = local_weights[pc % num_local_entries];
		for (i=0; i<local_history_length; i++,c++) {
			bool bit = !!(lh & (1<<i));

			// if the i'th bit in the local history is the same as
			// the current outcome, then increase, else decrease

			saturating_increment (c, taken == bit);
		}

		// train the global weights and coefficients

		for (i=0; i<history_length; i+=block_size) {
			int idx = index_entries(pc, ring_idx, i);
			byte *bl = &weights[i/block_size][idx * block_size];
			for (int j=0; j<block_size; j++) {
				int h = lut[index_history(ring_idx,i+j)];
				double partial_sum = h * bl[j];
				if ((partial_sum >= 0) == taken) {
					set_coeff (pc, i+j+1, get_coeff (pc, i+j+1) * factor);
				} else {
					set_coeff (pc, i+j+1, get_coeff (pc, i+j+1) / factor);
				}
				byte *c = bl + j;
				saturating_increment (c, taken == (bool) index_history (ring_idx, i+j));
			} 
		}
	}

	// get a piece of the local history into an unsigned int
	// so we can index the gshare predictor

        unsigned int get_history (int ring_idx, int n) {
                unsigned int x = 0;
                for (int i=0; i<n; i++) {
                        x <<= 1;
                        x |= index_history (ring_idx, i);
                }
                return x;
        }

	// predict the branch at this address 

	oh_snap_update *predict (unsigned int address, bool cond, oh_snap_update *u) { 
		u->ring_idx = spec_history_ring_idx;

		// only do this if we're predicting a conditional branch;
		// otherwise we'll just take a slot in the branch queue

		if (cond) {

			// shift off lower bit of address that contains no information

			address >>= 1;

			// find (or allocate) this branch in branch cache

			bc_entry *e = lookup (address);

			// if this branch has gone both ways before, predict it with the neural predictor

			if (e->ever_taken && e->ever_nottaken) {

				// get the gshare prediction

				byte *c = &gshare[(((get_history (spec_history_ring_idx, hl1) & hl1_mask) << hl1_shift) ^ address) % HYBRID_SIZE];
				bool small_prediction = (*c >= 2);

				// record the history

				u->lh = local_histories[address % num_local_histories];

				// if the global prediction is weak, use the local prediction
				// why make a half-assed predictor like this instead of a full-assed tournament-style predictor?
				// tried that; this one happens to work better!

				if ((*c == 2) || (*c == 1))
					small_prediction = local[(address ^ ((hl2_mask & u->lh) << hl2_shift)) & (HYBRID_SIZE-1)];

				// compute the neural output

				double yout = compute_output (address, u->lh, spec_history_ring_idx, e);

				// if this output is too weak, use the table-based predictor

				if (absolute_value (yout) < (threshold*theta[address % num_thetas])) {
					u->prediction (small_prediction);
				} else {
					u->prediction (yout >= 0.0);
				}
			} else {
				u->prediction (e->ever_taken);
			}
		} else {
			u->prediction (true);
		}
		return u;
	}

	// "speculatively" update the branch histories

	void spec_update (oh_snap_update *u, unsigned int address, bool taken) {
		address >>= 1;
		shift_history_ring (&spec_history_ring_idx, taken, address);

		// update local history

		unsigned int lh = local_histories[address % num_local_histories];
		lh <<= 1;
		lh |= taken;
		local_histories[address % num_local_histories] = lh;
	}

	// update an executed branch

	void update (oh_snap_update *u, unsigned int address, bool taken, bool cond) {
		address >>= 1;

		// only update for a conditional branch

		if (cond) {	

			// find the branch cache entry

			bc_entry *e = lookup (address);

			// update the gshare and local table-based predictors

			byte *c = &gshare[(((get_history (u->ring_idx, hl1) & hl1_mask) << hl1_shift) ^ address) % HYBRID_SIZE];
			if (taken) {
				if (*c < 3) (*c)++;
			} else {
				if (*c > 0) (*c)--;
			}
			local[(address ^ ((hl2_mask & u->lh) << hl2_shift)) & (HYBRID_SIZE-1)] = taken;

			// make a note of whether the branch has ever been take or not taken

			if (taken)
				e->ever_taken = true;
			else
				e->ever_nottaken = true;

			// if it's gone both ways, train the neural predictor

			if (e->ever_taken && e->ever_nottaken) {
				train (address, u->lh, u->ring_idx, e, taken, taken == u->prediction());
			} 
		}
	}

	// return the number of bits in the predictor

	int size (void) { 
		return size (block_size, num_entries, history_length, abits, bc_sets, bc_assoc, bc_tagbits, num_coefficients_vectors, coefficient_bits, local_history_length, num_local_histories, num_local_entries, num_thetas);
	}

	// static method to return the number of bits in the predictor
	static int size (int block_size, int num_entries, int history_length, int abits, int bcs, int bca, int bct, int num_coefficients_vectors, int coefficient_bits, int local_history_length, int num_local_histories, int num_local_entries, int num_thetas) {
		while (history_length % block_size) history_length++;

		// figure out organization of weights tables

		int i, k;
#if 0
		for (i=0; i<7; i++) nents[i] = 512;
		for (i=7; i<13; i++) nents[i] = 256;
		for (i=13; i<history_length/block_size+1; i++) nents[i] = 128;
#endif

		// figure out total number of bits used

		int bits = 0;

		// bits for in-flight history buffer; 9 bits for ring_idx
		// and 1 bit for prediction
		// and a local history 

		int branch_queue_bits = MAX_INFLIGHT_BRANCHES * (9 + 1 + local_history_length);
		printf ("branch queue bits = %d * (9 + 1 + %d) = %d\n", MAX_INFLIGHT_BRANCHES, local_history_length, branch_queue_bits);
		bits += branch_queue_bits;

		// bits for storing local histories

		int local_history_bits = num_local_histories * local_history_length;
		printf ("local history bits = %d * %d = %d\n", num_local_histories, local_history_length, local_history_bits);
		bits += local_history_bits;

		// bits for local history weights

		int local_perceptron_weights = WEIGHT_BITS * num_local_entries * local_history_length;
		printf ("local perceptron weights = %d * %d * %d = %d\n", WEIGHT_BITS, num_local_entries, local_history_length, local_perceptron_weights);
		bits += local_perceptron_weights;

		// bits used in weights tables

		int total_weights_bits = 0;
		for (i=0,k=0; i<history_length; i+=block_size,k++) {
			int blockk_bits = 
				  block_size /* weights in this block */
				* WEIGHT_BITS /* bits per weight */
				* nents[k] /* entries */;
			printf ("block %d bits = %d * %d * %d = %d\n", k, block_size, WEIGHT_BITS, nents[k], blockk_bits);
			total_weights_bits += blockk_bits;
		}
		printf ("total global weights bits = %d\n", total_weights_bits);
		bits += total_weights_bits;

		// bits used in branch cache

		int branch_cache_bits =
			 bcs * bca * 			// sets times assoc
			(bct + WEIGHT_BITS + 2); 	// tag bits + bias weight + taken/nottaken bits 
		printf ("branch cache bits = %d * %d * (%d + %d + 2) = %d\n", bcs, bca, bct, WEIGHT_BITS, branch_cache_bits);
		bits += branch_cache_bits;

		// bits used in little gshare

		int other_predictor_bits = HYBRID_SIZE * 3;
		printf ("other predictor bits = %d * 3 = %d\n", HYBRID_SIZE, other_predictor_bits);
		bits += other_predictor_bits;

		// bits used in coefficients vectors

		int coefficient_vector_bits = num_coefficients_vectors * coefficient_bits * (history_length+1);
		printf ("coefficient vector bits = %d * %d * (%d + 1) = %d\n", num_coefficients_vectors, coefficient_bits, history_length, coefficient_vector_bits);
		bits += coefficient_vector_bits;

		// bits used in pattern history vector

		int pattern_history_bits = history_length + MAX_INFLIGHT_BRANCHES;
		printf ("pattern history bits = %d + %d = %d\n", history_length, MAX_INFLIGHT_BRANCHES, pattern_history_bits);
		bits += pattern_history_bits;

		// bits used in path history vector

		int path_history_bits = abits * (history_length + MAX_INFLIGHT_BRANCHES);
		printf ("path history bits = %d * (%d + %d) = %d\n", abits, history_length, MAX_INFLIGHT_BRANCHES, path_history_bits);
		bits += path_history_bits;

		// bits used for thetas (they up to are 12 bits)

		int theta_bits = 12 * num_thetas;
		printf ("theta bits = 12 * %d = %d\n", num_thetas, theta_bits);
		bits += theta_bits;

		fflush (stdout);
		return bits;
	}
};

// insert a branch into the inflight branch state queue

oh_snap_update *insert_branch (unsigned int pc) {
	oh_snap_update *state = &inflight_branch_queue[queue_tail];
	queue_tail = (queue_tail + 1) % MAX_INFLIGHT_BRANCHES;
	return state;
}

// dequeue a branch

oh_snap_update *remove_branch (unsigned int pc) {
	oh_snap_update *r = &inflight_branch_queue[queue_head];
	queue_head = (queue_head + 1) % MAX_INFLIGHT_BRANCHES;
	return r;
}

// pointer to the predictor object

oh_snap *pred = NULL;

void PredictorInit() {
	pred = new oh_snap ();
	//printf ("predictor size is %d bytes\n", pred->size() / 8);
}

void PredictorReset() { }

// return true if this branch should be recorded in the history

bool is_branch (uint16_t type) {
	if (type & IS_BR_CONDITIONAL) return true;
	if (type & IS_BR_INDIRECT) return true;
	if (type & IS_BR_RETURN) return true;
	if (type & IS_BR_OTHER) return true;

	// IS_BR_CALL explicitly excluded

	if (0) if (type & IS_BR_CALL) return true;
	return false;
}

void PredictorRunACycle() {
	const cbp3_cycle_activity_t *cycle_info = get_cycle_info();

	for (int i = 0; i < cycle_info->num_fetch; i++) {
		uint32_t fe_ptr = cycle_info->fetch_q[i];
		const cbp3_uop_dynamic_t *uop = &fetch_entry(fe_ptr)->uop;
		if (is_branch (uop->type)) {
			bool cond = !!(uop->type & IS_BR_CONDITIONAL);

			// make a prediction with no knowledge of the actual outcome...

			oh_snap_update *state = pred->predict (uop->pc, cond, insert_branch (uop->pc));

			// report the prediction...

			if (cond) assert (report_pred(fe_ptr, false, state->prediction()));

			// now udpate the history "speculatively" with the
			// real outcome, after the prediction has been reported

			pred->spec_update (state, uop->pc, uop->br_taken);
		}
	}

	// update the predictor at retire stage

	for (int i = 0; i < cycle_info->num_retire; i++) {
		uint32_t rob_ptr = cycle_info->retire_q[i];
		const cbp3_uop_dynamic_t *uop = &rob_entry(rob_ptr)->uop;
		if (is_branch (uop->type)) {
			bool cond = !!(uop->type & IS_BR_CONDITIONAL);
			oh_snap_update *state = (oh_snap_update *) remove_branch (uop->pc);
			pred->update (state, uop->pc, uop->br_taken, cond);
		}
	}
}

void PredictorRunEnd() {
	rewind_marked = false;
}

void PredictorExit() { }
